/* idocument.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

public enum Valide.DocumentState
{
  NORMAL,
  EXTERNALLY_MODIFIED,
  BAD_ENCODING,
  INFO
}

/**
 * An interface for document
 */
public interface Valide.iDocument : Gtk.Widget
{
  /**
   * The signal savec is emited when the document is saved
   */
  public signal void saved ();

  /**
   * The signal savec is emited when the document is closed
   */
  public signal void closed ();

  /**
   * The signal savec is emited when the title of the document changed
   */
  public signal void title_change ();

  /**
   * The signal savec is emited when the state of the document changed
   */
  public signal void state_change (DocumentState state);

  /**
   * The state of the document
   */
  public abstract DocumentState state
  {
    get;
    set;
  }

  /**
   * The document is uptodate with the hard disk copy?
   */
  public abstract bool is_save
  {
    get;
    set;
  }

  /**
   * The absolute path of the document
   */
  public abstract string path
  {
    get;
    set;
  }

  /**
   * The title of the document (read-only)
   */
  public abstract string title
  {
    get;
  }

  /**
   * The base name of the file (read-only)
   */
  public abstract string filename
  {
    get;
  }

  /**
   * Open a document
   *
   * @param path A path of a file
   */
  public virtual void open (string path) throws Error
  {
    debug ("The class %s doesn't implement the iDocument.open function",
           this.get_type ().name ());
  }

  /**
   * Save the content of the widget in hard file file.
   * If the file has not yet been saved, this function show a dialog for enter
   * the path of the file.
   *
   * @param current_folder The current folder of chooser
   */
  public virtual void save (string? current_folder = null)
  {
    debug ("The class %s doesn't implement the iDocument.save function",
           this.get_type ().name ());
  }

  /**
   * Save as the content of the widget in hard disk file
   *
   * @param current_folder The current folder of chooser
   */
  public virtual void save_as (string? current_folder = null)
  {
    debug ("The class %s doesn't implement the iDocument.save_as function",
           this.get_type ().name ());
  }

  /**
   * Rename the hard disk file
   */
  public virtual void rename ()
  {
    debug ("The class %s doesn't implement the iDocument.rename function",
           this.get_type ().name ());
  }

  /**
   * Close the document
   */
  public virtual bool close ()
  {
    debug ("The class %s doesn't implement the iDocument.close function",
           this.get_type ().name ());
    return false;
  }

  /**
   * Get the pixbuf of the file
   *
   * @return A Gdk.Pixbuf
   */
  public virtual Gdk.Pixbuf get_icon ()
  {
    debug ("The class %s doesn't implement the iDocument.get_icon function",
           this.get_type ().name ());
    return Utils.get_pixbuf_for_stock (Gtk.STOCK_FILE, Gtk.IconSize.MENU);
  }

  /**
   * Get the informations about this file
   *
   * @return An information text
   */
  public virtual string get_tooltips ()
  {
    debug ("The class %s doesn't implement the iDocument.get_tooltips function",
           this.get_type ().name ());
    return "";
  }

  /**
   * Update the user interface
   *
   * @param ui_manager The UI manager
   */
  public virtual void setup_ui (UIManager ui_manager)
  {
    debug ("The class %s doesn't implement the iDocument.setup_ui function",
           this.get_type ().name ());
  }
}

