/* completion.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

using Valide;

class Valencia.ScanInfo : Object {
    public ParseInfo parse_info;
    public Method? method;
    public int cursor_pos;
    
    public ScanInfo(Method? method, ParseInfo parse_info, int cursor_position) {
        this.method = method;
        this.parse_info = parse_info;
        cursor_pos = cursor_position;
    }
    
    public ScanInfo.empty() { }
    
    public Expression inner() { return parse_info.inner; }
    
    public Expression outer() { return parse_info.outer; }
}

public class Completion : Plugin, Object
{
  private uint ui_id;
  private Popup popup;
  private Calltips calltips;
  private SourceBuffer buffer;
  private Gtk.ActionGroup action_group;
  private SList<Pair<weak Object, ulong>> signal_handler_list;

  const Gtk.ActionEntry[] entries = {
        { "edit-goto-definition", null, "Go to Definition", "<ctrl>D",
          "Goto the definition of the current symbol", on_goto_definition },
        { "edit-autocomplete", null, "_AutoComplete", "<ctrl>Return",
          "Display method or symbol information", on_autocomplete }
    };

  const string ui = """
      <ui>
        <menubar name="menubar">
          <menu action="search">
            <placeholder name="search-ops-3">
              <menuitem action="edit-goto-definition"/>
              <separator/>
              <menuitem action="edit-autocomplete"/>
            </placeholder>
          </menu>
        </menubar>
      </ui>""";

  public string path { get; construct set; }
  public Window window { get; construct set; }

  private void get_window_position_in_cursor (Gtk.Window window,
                                              ref int x, ref int y)
  {
    int w, h;
    int sw, sh;
    int xtext, ytext;

    sw = Gdk.Screen.width ();
    sh = Gdk.Screen.height ();

    window.get_size (out w, out h);
    if (x + w > sw) x = sw - w - 4;
    // If we cannot show it down, we show it up.
    if (y + h > sh)
    {
      Pango.Layout layout;

      layout = window.create_pango_layout (null);
      layout.get_pixel_size (out xtext, out ytext);
      y = y - ytext - h;
    }
  }

  private void get_cursor_pos (Gtk.TextView text_view, 
                               out int x, out int y)
  {
    int xx, yy;
    Gdk.Window win;
    int win_x, win_y;
    Gtk.TextIter start;
    Gdk.Rectangle location;
    Gtk.TextMark insert_mark;
    Gtk.TextBuffer text_buffer;

    text_buffer = text_view.get_buffer ();
    insert_mark = text_buffer.get_insert ();
    text_buffer.get_iter_at_mark (out start,insert_mark);
    text_view.get_iter_location (start, out location);
    text_view.buffer_to_window_coords (Gtk.TextWindowType.WIDGET, location.x,
                                       location.y, out win_x, out win_y);

    win = text_view.get_window (Gtk.TextWindowType.WIDGET);
    win.get_origin (out xx, out yy);

    x = win_x + xx;
    y = win_y + yy + location.height;
  }

  private Gdk.Pixbuf valencia_type_to_icon (Valencia.Symbol symbol)
  {
    string type;
    string valencia_type;

    valencia_type = symbol.get_type ().name ();
    type = valencia_type.substring ("Valencia".len ()).down ();
    return Utils.get_symbol_pixbuf (type);
  }

  private void get_buffer_info (SourceBuffer buffer, out int pos,
                                out string contents)
  {
    Gtk.TextIter iter;

    contents = buffer.get_buffer_contents ();
    this.buffer.get_iter_at_mark (out iter, this.buffer.get_insert ());
    pos = iter.get_offset ();
  }

  Valencia.ScanInfo? get_scan_info (bool partial)
  {
    int cursor_pos;
    string source;
    Valencia.Method? method;
    Valencia.ParseInfo parse_info;

    this.get_buffer_info (this.buffer, out cursor_pos, out source);
    parse_info = new Valencia.ExpressionParser (source, cursor_pos, partial).parse ();

    Valencia.Program program = Valencia.Program.find_containing (this.buffer.path);
    Valencia.SourceFile sf = program.find_source(this.buffer.path);
    // The sourcefile may be null if the file is a vala file but hasn't been saved to disk
    if (sf == null)
    {
      return null;
    }

    // Give the method tooltip precedence over autocomplete
    method = null;
    if (parse_info.outer != null && 
        (!this.calltips.visible /*|| cursor_is_inside_different_function(parse_info.outer_pos)*/))
    {
      Valencia.Symbol? sym;

      sym = sf.resolve (parse_info.outer, cursor_pos, false);
      if (sym != null)
      {
        method = sym as Valencia.Method;
      }
    }

    return new Valencia.ScanInfo (method, parse_info, cursor_pos);
  }

  private void on_parse_end ()
  {
    Valencia.ScanInfo? info;

    info = this.get_scan_info (true);
    if (info != null)
    {
      if (info.method != null)
      {
        /* calltips */
        this.calltips.text = info.method.to_string ();
        this.calltips.show ();
      }
      else
      {
        /* popup */
        Valencia.Expression e;
        Valencia.SourceFile sf;
        Valencia.Program program;
        Valencia.SymbolSet symbol_set;
        Gee.HashSet<Valencia.Symbol>? symbols;

        e = info.inner ();
        if (e == null)
        {
          if (info.method != null)
          {
            return;
          }
          e = new Valencia.Id ("");
        }

        program = Valencia.Program.find_containing (this.buffer.path);
        sf = program.find_source (this.buffer.path);
        symbol_set = sf.resolve_prefix (e, info.cursor_pos, false);
        symbols = symbol_set.get_symbols ();
        if (symbols != null)
        {
          List<Proposal> proposals;

          proposals = new List <Proposal> ();
          foreach (Valencia.Symbol s in symbols)
          {
            proposals.append (new Proposal (s.name, this.valencia_type_to_icon (s)));
          }
          this.popup.proposals = proposals;
          this.popup.show ();
        }
        else
        {
          this.popup.clear ();
        }
      }
    }
  }

  private void hide_popup ()
  {
    this.popup.hide ();
    this.calltips.hide ();
  }

  private void show_popup ()
  {
    Valencia.Program program;

    program = Valencia.Program.find_containing (this.buffer.path, true);
    if (program.is_parsing ())
    {
      program.system_parse_complete.connect (this.on_parse_end);
    }
    else
    {
      this.on_parse_end ();
    }
  }

  private void on_autocomplete ()
  {
    iDocument document;

    document = this.window.documents.current;
    if (document is AbstractTextDocument)
    {
      int x, y;
      SourceView source_view;
      AbstractTextDocument atd;

      atd = document as AbstractTextDocument;
      source_view = atd.text_view;
      this.get_cursor_pos (source_view, out x, out y);
      this.get_window_position_in_cursor (this.popup, ref x, ref y);
      this.popup.move (x, y);
      this.calltips.move (x, y);
      this.buffer = source_view.buffer;
      this.show_popup ();
    }
  }

  private void jump_to_symbol_definition ()
  {
    int pos;
    string source;
    Valencia.ScanInfo info;

    this.get_buffer_info (this.buffer, out pos, out source);
    info = get_scan_info (false);
    if (info != null && info.inner () != null)
    {
      Valencia.Symbol? sym;
      Valencia.SourceFile sf;
      Valencia.Program program;

      program = Valencia.Program.find_containing (this.buffer.path);
      sf = program.find_source (this.buffer.path);
      sym = sf.resolve (info.inner (), pos, false);
      if (sym != null)
      {
        iDocument document;

        document = this.window.documents.create (sym.source.filename);
        if (document is AbstractTextDocument)
        {
          AbstractTextDocument atd;

          atd = document as AbstractTextDocument;
          atd.select_offsets (sym.start, sym.start + sym.name_length ());
        }
      }
    }
  }
      
  private void on_goto_definition ()
  {
    iDocument document;

    document = this.window.documents.current;
    if (document is AbstractTextDocument)
    {
      string filename;
      SourceView source_view;
      AbstractTextDocument atd;

      atd = document as AbstractTextDocument;
      source_view = atd.text_view;
      filename = source_view.buffer.path;

      if (filename != null && Valencia.Program.is_vala (filename))
      {
        Valencia.Program program;

        program = Valencia.Program.find_containing (filename, true);
        this.buffer = source_view.buffer;
        if (program.is_parsing ())
        {
          program.system_parse_complete.connect (this.jump_to_symbol_definition);
        }
        else
        {
          this.jump_to_symbol_definition ();
        }
      }
    }
  }

  private static bool on_key_press (Gtk.Widget sender, Gdk.EventKey event,
                                    Completion self)
  {
    bool catched = false;

    if (self.popup.visible || self.calltips.visible)
    {
      switch (event.keyval)
      {
        case Gdk.KeySyms.Left:
        case Gdk.KeySyms.Right:
          catched = true;
        break;
        case Gdk.KeySyms.Down:
          self.popup.select_next (1);
          catched = true;
        break;
        case Gdk.KeySyms.Page_Down:
          self.popup.select_next (5);
          catched = true;
        break;
        case Gdk.KeySyms.Up:
          self.popup.select_previous (1);
          catched = true;
        break;
        case Gdk.KeySyms.Page_Up:
          self.popup.select_previous (5);
          catched = true;
        break;
        case Gdk.KeySyms.Home:
          self.popup.select_first ();
          catched = true;
        break;
        case Gdk.KeySyms.End:
          self.popup.select_last ();
          catched = true;
        break;
        case Gdk.KeySyms.Return:
        case Gdk.KeySyms.Tab:
          Proposal prop = null;

          if (self.popup.get_selected_proposal (out prop))
          {
            self.buffer.replace_actual_word (prop.label);
            self.popup.hide ();
            catched = true;
          }
        break;
        case Gdk.KeySyms.Escape:
          self.hide_popup ();
          catched = true;
        break;
      }
    }
    return catched;
  }

  private static void on_delete_range (Gtk.TextBuffer sender, Gtk.TextIter start,
                                       Gtk.TextIter end, Completion self)
  {
    if (self.popup.visible || self.calltips.visible)
    {
      self.hide_popup ();
      self.show_popup ();
    }
  }

  private static void on_insert_text (Gtk.TextBuffer sender, Gtk.TextIter iter,
                                      string text, int length, Completion self)
  {
    if (self.popup.visible)
    {
      if (text.get_char ().isspace ())
      {
        self.popup.hide ();
      }
      else if (text == "(")
      {
        self.popup.hide ();
        self.show_popup ();
      }
      else
      {
        self.show_popup ();
      }
    }

    if (self.calltips.visible)
    {
      if (text == ")")
      {
        self.calltips.hide ();
      }
    }
  }

  private static bool on_button_press (Gtk.Widget sender, Gdk.EventButton event,
                                       Completion self)
  {
    self.hide_popup ();
    return false;
  }

  private static bool on_focus_out  (Gtk.Widget sender, Gdk.EventFocus event,
                                     Completion self)
  {
    self.hide_popup ();
    return false;
  }

  private static void on_tab_added (DocumentManager documents,
                                    iDocument document, Completion self)
  {
    if (document is AbstractTextDocument)
    {
      SourceBuffer buffer;
      AbstractTextDocument atd;

      atd = (document as AbstractTextDocument);
      buffer = atd.buffer;

      self.add_signal (atd.text_view, "key_press_event",
                       (Callback)self.on_key_press);
      self.add_signal (atd.text_view, "focus_out_event",
                       (Callback)self.on_focus_out);
      self.add_signal (atd.text_view, "button_press_event",
                       (Callback)self.on_button_press);
      self.add_signal (atd.buffer, "insert_text",
                       (Callback)self.on_insert_text, true);
      self.add_signal (atd.buffer, "delete_range",
                       (Callback)self.on_delete_range, true);
    }
  }

  private static void setup_ui (DocumentManager documents, iDocument document,
                         Completion self)
  {
    bool active = false;

    if (documents.current != null && documents.current is AbstractTextDocument)
    {
      active = true;
    }
    self.action_group.get_action ("edit-goto-definition").sensitive = active;
    self.action_group.get_action ("edit-autocomplete").sensitive = active;
  }

  private void add_signal (Object object, string signal_name, Callback cb, 
                           bool after = false)
  {
    ulong id;

    if (!after)
    {
      id = Signal.connect (object, signal_name, cb, this);
    }
    else
    {
      id = Signal.connect_after (object, signal_name, cb, this);
    }
    this.signal_handler_list.append (new Pair<Object, ulong> (object, id));
  }

  construct
  {
    Gtk.UIManager ui_manager;

    this.popup = new Popup ();
    this.calltips = new Calltips ();
    this.signal_handler_list = new SList<Pair<weak Object, ulong>> ();

    this.add_signal (this.window.documents, "tab_added",
                     (Callback)this.on_tab_added);
    this.add_signal (this.window.documents, "tab_removed",
                     (Callback)this.setup_ui);
    this.add_signal (this.window.documents, "tab_changed",
                     (Callback)this.setup_ui);

    /* Setup iu */
    ui_manager = this.window.ui_manager;
    this.action_group = new Gtk.ActionGroup ("completion");
    this.action_group.add_actions (this.entries, this);
    ui_manager.insert_action_group (action_group, 0);
    this.ui_id = ui_manager.add_ui_from_string(ui, -1);
  }

  ~Completion ()
  {
    this.window.ui_manager.remove_ui (this.ui_id);

    foreach (Pair<Object, ulong> pair in this.signal_handler_list)
    {
      if (SignalHandler.is_connected (pair.first, pair.second))
      {
        SignalHandler.disconnect (pair.first, pair.second);
      }
    }
  }
}

[ModuleInit]
public Type register_plugin ()
{
  return typeof (Completion);
}

