/* symbol-browser.vala
 *
 * Copyright (C) 2008-2009  Nicolas Joseph
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 * 	Nicolas Joseph <nicolas.joseph@valaide.org>
 */

/*
 * This code is re-writing in the vala language of GeditSymbolBrowserPlugin
 *
 * Copyright (C) 2007 - Micah Carrick
 */

using Valide;

/* just used to keep the population code readable */
private class Symbol
{
  public string name;
  public string full_name;
  public string signature;
  public string[] inherits;
  public string typeref;
  public string access;
  public string implementation;
  public string type;
  public string file;
  public string language;
  public int line;
  public Symbol root;
  public weak AbstractTextDocument document;
}

private errordomain SymbolError
{
  KIND
}

public class SymbolBrowser : Plugin, Object
{
  /* enum for tree model columns, not all is currently used */
  enum Col
  {
    ICON,
    NAME,
    FILE,
    LINE,
    LANGUAGE,
    NB
  }

  private Gtk.VBox box;
  private Gtk.TreeView tree_view;
  private Gtk.TreeStore model;
  private Gtk.Label label;
  private SList<string> expanded_rows;
  private uint status_context;
  private Gtk.TreeViewColumn language_column;
  private Gtk.TreeViewColumn source_file_column;
  private Gtk.TreeViewColumn line_number_column;
  private HashTable<string, Gtk.TreeIter?> root_nodes;
  private int atom;

  public string path { get; construct set; }
  public Window window { get; construct set; }

  private bool show_language_column
  {
    get
    {
      return this.window.config_manager.get_boolean ("SymbolBrowser", "show-language");
    }
    set
    {
      this.window.config_manager.set_boolean ("SymbolBrowser", "show-language",
                                                 value);
      this.language_column.visible = value;
    }
  }

  private bool show_source_file_column
  {
    get
    {
      return this.window.config_manager.get_boolean ("SymbolBrowser", "show-source-file");
    }
    set
    {
      this.window.config_manager.set_boolean ("SymbolBrowser", "show-source-file",
                                                 value);
      this.source_file_column.visible = value;
    }
  }

  private bool show_line_number_column
  {
    get
    {
      return this.window.config_manager.get_boolean ("SymbolBrowser", "show-line-number");
    }
    set
    {
      this.window.config_manager.set_boolean ("SymbolBrowser", "show-line-number",
                                                 value);
      this.line_number_column.visible = value;
    }
  }

  private bool only_load_active_tab
  {
    get
    {
      return true;
    }
  }

  private bool hierarchic_view
  {
    get
    {
      return this.window.config_manager.get_boolean ("SymbolBrowser", "hierarchic-view");
    }
    set
    {
      this.window.config_manager.set_boolean ("SymbolBrowser", "hierarchic-view", value);
      this.populate_symbols ();
    }
  }

  private void save_expanded_rows ()
  {
/*
        GConfClient             *client;
        GError                  *error=NULL;
        
        client = gconf_client_get_default ();
        
        gconf_client_set_list (client, 
                        GEDIT_SYMBOL_BROWSER_PLUGIN_GCONF_PATH"expanded-rows",
                        GCONF_VALUE_STRING, panel->priv->expanded_rows,
                        &error);
                        
        if (error != NULL)
        {
                g_warning ("Error saving gedit-symbol-browser-plugin setting: %s", 
                                error->message);
                                
                g_error_free (error);
        }

        g_object_unref (client);
*/
  }

  private bool is_expanded_row (string name)
  {
    bool ret = false;

    foreach (string i in this.expanded_rows)
    {
      if (i == name)
      {
        ret = true;
        break;
      }
    }
    return ret;
  }

  private void add_expanded_row (string name)
  {
    if (!this.is_expanded_row (name))
    {
      this.expanded_rows.prepend (name);
      this.save_expanded_rows ();
    }
  }

  private void remove_expanded_row (string name)
  {
    weak SList<string> l;

    if (this.is_expanded_row (name) == true)
    {
      for (l = this.expanded_rows; l != null; l = l.next)
      {
        if (l.data == name)
        {
          this.expanded_rows.remove_all (l.data);
        }
      }
      this.save_expanded_rows ();
    }
  }

  /**
   * Execute the ctags command and return stdout.
   */ 
  private string exec_ctags (string filename) throws Error
  {
    string output;
    string command;

    /* build ctags command */
    command = "%s --extra=+q --fields=afmiKlnsStz -f - \"%s\""
                       .printf (Config.VALIDE_CTAGS_EXEC, filename);
    /* execute command */
    Process.spawn_command_line_sync (command, out output, null, null);

    return output;
  }

  /**
   * This function needs work. Used to be custom parsed for ex searches
   * which could potentially contain tabs, however, I've decided that isn't
   * really needed. But, it's still using Symbol. We could just pass back
   * **fields array.
   */ 
  private Symbol? parse_line (string line)
  {
    /* (parent.)name file ex_cmd;" kind:[enumvalue|errorcode|errordomain|enum|struct|namespace] line:x language:x access:x */
    /* (parent.)name file ex_cmd;" kind:field line:x language:x typeref:x access:x */
    /* (parent.)name file ex_cmd;" kind:class line:x language:x inherits:x access:x implementation:x */
    /* (parent.)name file ex_cmd;" kind:[delegate|method|signal] line:x language:x typeref:x access:x implementation:x signature:x */
    /* (parent.)name file ex_cmd;" kind:interface line:x language:x inherits:x access:x */
    /* (parent.)name file ex_cmd;" kind:property line:x language:x typeref:x access:x implementation:x */
    /* name          file	ex_cmd;" kind:function line:x language:x signature:x */
    /* name          file ex_cmd;" kind:variable line:x language:x */

    string[] fields;
    Symbol symbol = null;

    fields = line.split ("\t", 0);
    if (fields[0] != null)
    {
      string filename;
      string name;
      string full_name;
      string kind;
      int file_position;
      string language;
      string access = null;
      string[] inherits = null;
      string typeref = null;
      string signature = null;
      string implementation = null;
      string[] splits;

      symbol = new Symbol ();
      full_name = fields[0];
      if (full_name.rchr (-1, '.') != null)
      {
        name = full_name.rchr (-1, '.').offset (1);
      }
      else
      {
        name = full_name;
      }
      filename = fields[1];
      splits = fields[3].split(":", 0);
      kind = splits[1];
      if (kind == null)
      {
        return null;
      }

      splits = fields[4].split(":", 0);
      file_position = splits[1].to_int ();

      splits = fields[5].split(":", 0);
      language = fields[1];

      switch (kind)
      {
        case "enumvalue":
        case "errorcode":
        case "errordomain":
        case "enum":
        case "namespace":
        case "struct":
          splits = fields[5].split(":", 0);
          access = splits[1];
        break;
        case "field":
          splits = fields[6].split(":", 0);
          access = splits[1];
        break;
        case "class":
          splits = fields[6].split(":", 0);
          if (splits != null && splits[1] != null)
          {
            inherits = splits[1].split (",", 0);
          }
          splits = fields[7].split(":", 0);
          access = splits[1];
          splits = fields[8].split(":", 0);
          implementation = splits[1];
        break;
        case "delegate":
        case "method":
        case "signal":
          splits = fields[6].split(":", 0);
          access = splits[1];
          splits = fields[7].split(":", 0);
          implementation = splits[1];
          splits = fields[8].split(":", 0);
          signature = splits[1];
        break;
        case "interface":
          splits = fields[6].split(":", 0);
          if (splits != null && splits[1] != null)
          {
            inherits = splits[1].split (",", 0);
          }
          splits = fields[7].split(":", 0);
          access = splits[1];
        break;
        case "property":
          splits = fields[6].split(":", 0);
          access = splits[1];
          splits = fields[7].split(":", 0);
          implementation = splits[1];
        break;
        case "function":
          splits = fields[5].split(":", 0);
          signature = splits[1];
        break;
        case "variable":
        break;
        default:
          debug ("Unknow kind '%s' for symbol '%s'", kind, name);
        break;
      }

      symbol.name = name;
      symbol.full_name = full_name;
      symbol.file = filename;
      symbol.line = file_position;
      symbol.signature = signature;
      symbol.inherits = inherits;
      symbol.typeref = typeref;
      symbol.root = null;
      if (kind != null)
      {
        symbol.type = kind;
      }
      if (access != null)
      {
        symbol.access = access;
      }
      if (implementation != null)
      {
        symbol.implementation = implementation;
      }
    }
    return symbol;
  }

  private void clear_data ()
  {
    this.model.clear ();
    this.root_nodes = new HashTable<string, Gtk.TreeIter?> (str_hash, str_equal);
    this.box.sensitive = false;
    this.label.set_text ("");
  }

  /**
   * Add a single symbol to the treeview. Parents rows are added as the type
   * whenever they aren't already present.
   */
  private bool add_symbol (Symbol symbol)
  {
    string markup;
    Gdk.Pixbuf pixbuf;
    Gtk.TreeIter iter;
    Gtk.TreeIter? parent;

    Utils.process_gtk_events ();
    symbol.name = symbol.name.replace ("<", "&lt;");
    symbol.name = symbol.name.replace (">", "&gt;");

    pixbuf = Utils.get_symbol_pixbuf (symbol.type);

    if (!this.hierarchic_view)
    {
      /* we could do bold here */
      markup = "<b>%s</b>".printf (symbol.name);

      /* check for existing parent iter */
      parent = this.root_nodes.lookup (markup);
      /* create the parent if it wasn't already found */
      if (parent == null)
      {
        Gtk.TreeIter tmp;

        this.model.append (out tmp, null);
        parent = tmp;
        this.model.set (parent, Col.ICON, pixbuf, Col.NAME, markup);
        this.root_nodes.insert (markup, parent);
      }

      this.model.append (out iter, parent);
      /* add the new symbol */
      this.model.set (iter, Col.ICON, pixbuf,
                            Col.NAME, symbol.name,
                            Col.FILE, symbol.file,
                            Col.LINE, symbol.line,
                            Col.LANGUAGE, symbol.language);
    }
    else
    {
      if (symbol.root != null)
      {
        parent = this.root_nodes.lookup (symbol.root.full_name);
        this.model.append (out iter, parent);
      }
      else
      {
        this.model.append (out iter, null);
        parent = iter;
      }
      this.model.set (iter, Col.ICON, pixbuf,
                            Col.NAME, symbol.name,
                            Col.FILE, symbol.file,
                            Col.LINE, symbol.line,
                            Col.LANGUAGE, symbol.language);
      this.root_nodes.insert (symbol.full_name, iter);
    }
    return true;
  }

  private bool populate_symbols_asc ()
  {
    if (AtomicInt.compare_and_exchange (ref this.atom, 0, 1))
    {
      this.model = this.tree_view.get_model () as Gtk.TreeStore;
      this.tree_view.set_model (null);
      this.tree_view.freeze_child_notify ();
      if (this.window.documents.current != null)
      {
        List<iDocument> doc;

        doc = new List<iDocument> ();
        /* find out if we want to load all symbols or just active tab */
        if (this.only_load_active_tab)
        {
          doc.prepend (this.window.documents.current);
        }
        else
        {
          doc = this.window.documents.documents.copy ();
        }

        this.clear_data ();
        /* run ctags on each document in the list and add symbols */
        foreach (iDocument d in doc)
        {
          if (d is AbstractTextDocument && d.path != null
              && FileUtils.test (d.path, FileTest.EXISTS))
          {
            try
            {
              string output = this.exec_ctags (d.path);

              Symbol root = null;
              string[] lines = output.split ("\n", 0);
              for (int i = 0; lines[i] != null; i++)
              {
                if (lines[i] != "")
                {
                  Symbol symbol;

                  symbol = this.parse_line (lines[i]);
                  if (symbol != null)
                  {
                    symbol.document = (AbstractTextDocument)d;
                    if (symbol.type == "class")
                    {
                      root = null;
                    }
                    else if (root != null
                             && ((root.type == "enum" && symbol.type != "enumvalue")
                             || (root.type == "errordomain" && symbol.type != "errorcode")))
                    {
                      root = root.root;
                    }
                    symbol.root = root;
                    this.add_symbol (symbol);
                    if (symbol.type == "class" || symbol.type == "enum"
                        || symbol.type == "errordomain")
                    {
                      root = symbol;
                    }
                  }
                }
              }
            }
            catch (Error e)
            {
              debug (_("Could not execute ctags: %s\n"), e.message);
            }
          }
        }
        if (this.model.iter_n_children (null) > 0)
        {
          this.box.sensitive = true;
        }
      }
      else
      {
        this.clear_data ();
      }
      this.tree_view.set_model (this.model);
      this.tree_view.thaw_child_notify ();
//      this.restore_expanded_rows ();
      this.tree_view.expand_all ();
      AtomicInt.set (ref this.atom, 0);
    }
    return false;
  }

  /**
   * Call to re-populate tree view.
   */
  private void populate_symbols ()
  {
    Idle.add (this.populate_symbols_asc);
  }

  /**
   * Re-expand any row when the list is repopulated if that row was previously
   * expanded.
   */
  private void restore_expanded_rows ()
  {
    /* is this going to fix the crash? */
    if (this.window.documents.sensitive == true)
    {
      Gtk.TreeIter iter;
      Gtk.TreeModel model = this.tree_view.get_model ();

      if (model.get_iter_first (out iter) == true)
      {
        for (bool r = model.get_iter_first (out iter); 
             r == true;
             r = model.iter_next (ref iter))
        {
          string name;

          model.get (iter, Col.NAME, out name,  -1);
          if (this.is_expanded_row (name))
          {
            Gtk.TreePath path = model.get_path (iter);
            this.tree_view.expand_row (path, false);
          }
        }
      }
    }
  }

  private void on_tab_changed ()
  {
    /* only needed in single-file mode  */
    if (this.only_load_active_tab)
    {
      this.populate_symbols ();
    }
  }

  private void on_row_activated (Gtk.TreeView sender, Gtk.TreePath path,
                                 Gtk.TreeViewColumn column)
  {
    Gtk.TreeIter iter;
        
    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get_iter (out iter, path);

    /* expand/collapse row if it has children (and return) */
    if (!this.hierarchic_view && model.iter_has_child (iter))
    {
      if (this.tree_view.is_row_expanded (path))
      {
        this.tree_view.collapse_row (path);
      }
      else
      {
        this.tree_view.expand_row (path, false);
      }
    }
    else
    {
      /* otherwise, show the file name and line number and flash statusbar */
      int line;
      string name;
      string filename;

      model.get (iter, Col.NAME, out name,  Col.LINE, out line,
                 Col.FILE, out filename);
      try
      {
        iDocument document;

        document = this.window.documents.create (filename);
        if (document is AbstractTextDocument)
        {
          (document as AbstractTextDocument).text_view.goto_line (line - 1);
        }
        document.grab_focus ();
      }
      catch (Error e)
      {
        warning (e.message);
      }
    }
  }

  private void on_cursor_changed (Gtk.TreeView sender)
  {
    Gtk.TreeIter iter;
    Gtk.TreeModel model;
    Gtk.TreeSelection selection;

    model = this.tree_view.get_model ();
    selection = this.tree_view.get_selection ();
    if (selection.get_selected (null, out iter))
    {
      int line;
      string name;
      string filename;

      model.get (iter, Col.NAME, out name, Col.FILE, out filename,
                       Col.LINE, out line);

			if (filename != null)
			{
		    string label_text;
		    string status_message;

		    label_text = "%s Line:%d".printf (Path.get_basename (filename), line);
		    this.label.set_text (label_text);

		    status_message = "Double-click to jump to '%s'".printf (name);
		    this.window.statusbar.flash_message (this.status_context, status_message);
		  }
    }
  }

  private void on_row_expanded (Gtk.TreeView sender, Gtk.TreeIter iter,
                                Gtk.TreePath path)
  {
    string name;        

    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get (iter, Col.NAME, out name, -1);
    return_if_fail (name != null);

    this.add_expanded_row (name);
  }

  private void on_row_collapsed (Gtk.TreeView sender, Gtk.TreeIter iter,
                                 Gtk.TreePath path)
  {
    string name;

    Gtk.TreeModel model = this.tree_view.get_model ();
    model.get (iter, Col.NAME, out name, -1);
    return_if_fail (name != null);

    remove_expanded_row (name);
  }

  construct
  {
    Gtk.ScrolledWindow scrolled;
    Gtk.TreeStore store;
    Gtk.TreeViewColumn col;
    Gtk.CellRendererText renderer;
    Gtk.CellRendererPixbuf renderer_pixbuf;
    Gtk.Image image;

    AtomicInt.set (ref this.atom, 0);

    this.box = new Gtk.VBox (false, 0);

    /* create some nice padding */
    this.box.set_spacing (2);
    this.box.set_border_width (2);
 
    /* create and pack tree view */
    scrolled = new Gtk.ScrolledWindow (null, null);
    scrolled.set_policy (Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
    scrolled.set_shadow_type (Gtk.ShadowType.IN);
    this.box.pack_start (scrolled, true, true, 0);

    this.tree_view = new Gtk.TreeView ();
    this.tree_view.set_rules_hint (false);
    this.tree_view.set_headers_visible (true);
    this.tree_view.set_headers_clickable (true);
    scrolled.add (this.tree_view);

    /* create and pack label */
    this.label = new Gtk.Label ("");
    this.label.set_alignment ((float)0.0, (float)0.5);
    this.box.pack_start (this.label, false, true, 0);

    /* create model */
    store = new Gtk.TreeStore (Col.NB, typeof (Gdk.Pixbuf), typeof (string),
                               typeof (string), typeof (int), typeof (string),
                               typeof (Object));
    //store.set_sort_column_id (1, Gtk.SortType.ASCENDING);

    /* main column */
    col = new Gtk.TreeViewColumn ();
    col.set_title(_("Symbol"));

    renderer_pixbuf = new Gtk.CellRendererPixbuf ();
    col.pack_start (renderer_pixbuf, false);
    col.set_attributes (renderer_pixbuf, "pixbuf", Col.ICON);

    renderer = new Gtk.CellRendererText ();
    col.pack_start (renderer, false);
    col.set_attributes (renderer, "markup", Col.NAME);
    this.tree_view.append_column (col);

    /* additional columns */

    this.language_column = new Gtk.TreeViewColumn ();
    this.show_language_column = this.window.config_manager.get_boolean ("SymbolBrowser", "show-language");
    this.language_column.set_title (_("Language"));
    renderer = new Gtk.CellRendererText ();
    this.language_column.pack_start (renderer, true);
    this.language_column.set_attributes (renderer, "text", Col.LANGUAGE);
    this.tree_view.append_column (this.language_column);

    this.source_file_column = new Gtk.TreeViewColumn ();
    this.show_source_file_column = this.window.config_manager.get_boolean ("SymbolBrowser", "show-source-file");
    this.source_file_column.set_title (_("File"));
    renderer = new Gtk.CellRendererText ();
    this.source_file_column.pack_start (renderer, true);
    this.source_file_column.set_attributes (renderer, "text", Col.FILE);
    this.tree_view.append_column (this.source_file_column);

    this.line_number_column = new Gtk.TreeViewColumn ();
    this.show_line_number_column = this.window.config_manager.get_boolean ("SymbolBrowser", "show-line-number");
    this.line_number_column.set_title (_("Line"));
    renderer = new Gtk.CellRendererText ();
    this.line_number_column.pack_start (renderer, true);
    this.line_number_column.set_attributes (renderer, "text", Col.LINE);
    this.tree_view.append_column (this.line_number_column);

    this.tree_view.set_model (store);

    /* attach signal callbacks for tree view */
    this.tree_view.row_activated.connect (on_row_activated);
    this.tree_view.cursor_changed.connect (on_cursor_changed);
    this.tree_view.row_expanded.connect (on_row_expanded);
    this.tree_view.row_collapsed.connect (on_row_collapsed);

    this.status_context = this.window.statusbar.get_context_id ("valide_symbol_browser");

    this.box.sensitive = false;

    this.window.documents.tab_removed.connect (this.populate_symbols);
    this.window.documents.tab_changed.connect (this.on_tab_changed);
    this.window.documents.tab_saved.connect (this.on_tab_changed);

    image = new Gtk.Image.from_file (Path.build_filename (this.path,
                                                          "symbol-browser.png"));

    this.window.side_panel.add_item (this.box, "Symbol Browser", image);

    this.box.show_all ();
  }

  public SymbolBrowser (string path)
  {
    this.path = path;
  }

  ~SymbolBrowser ()
  {
    this.window.documents.tab_removed.disconnect (this.populate_symbols);
    this.window.documents.tab_changed.disconnect (this.on_tab_changed);
    this.window.documents.tab_saved.connect (this.on_tab_changed);
    this.window.side_panel.remove_item (this.box);
  }

	public Gtk.Widget create_configure_dialog ()
	{
	  Gtk.Frame frame;
	  Gtk.VBox vbox;
	  Gtk.CheckButton check;

	  frame = new Gtk.Frame (_("Additionnal columns"));

	  vbox = new Gtk.VBox (true, 0);
	  frame.add (vbox);

	  check = new Gtk.CheckButton.with_label (_("Programming language"));
	  check.set_active (this.show_language_column);
	  check.toggled.connect ((s) => {
	    this.show_language_column = s.get_active ();
	  });
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Source file"));
	  check.set_active (this.show_source_file_column);
	  check.toggled.connect ((s) => {
	    this.show_source_file_column = s.get_active ();
	  });
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Line number"));
	  check.set_active (this.show_line_number_column);
	  check.toggled.connect ((s) => {
	    this.show_line_number_column = s.get_active ();
	  });
	  vbox.pack_start (check, true, true, 0);

	  check = new Gtk.CheckButton.with_label (_("Hierarchic view"));
	  check.set_active (this.hierarchic_view);
	  check.toggled.connect ((s) => {
	    this.hierarchic_view = s.get_active ();
	  });
	  vbox.pack_start (check, true, true, 0);

	  return frame;
	}
}

[ModuleInit]
public Type register_plugin ()
{
  return typeof (SymbolBrowser);
}

public Gtk.Widget create_configure_dialog (SymbolBrowser self)
{
  return self.create_configure_dialog ();
}

